@@ -7,7 +7,8 @@
     Challenges:
     1. config files are hierarchical, args & envs are flat
     2. cli args and config keys have dashes, envvars have underscores
-    3. some input is intended to be interactive-only - such as --version, --help, --long-help
+    3. cli args have boolean flags, envvars have strings, config files have boolean values
+    4. some input is intended to be interactive-only - such as --version, --help, --long-help
 """
 
 # type: ignore  # pylint: disable=no-member
@@ -73,7 +74,7 @@ VALID_CONFIG_PROP_VALUES['action'] = ['store_const', 'store_true', 'store_false'
 STANDARD_CONFIGS: META_CONFIG_TYPE = {}
 # IO Config Items:
 STANDARD_CONFIGS['infiles'] = {'short_name': 'i',
-                               'default': '-',
+                               'default': ['-'],
                                'required': True,
                                'type': str,
                                'nargs': '*'}
@@ -134,7 +135,8 @@ STANDARD_CONFIGS['verbosity'] = {'default': 'normal',
 
 STANDARD_CONFIGS['dry_run'] = {'default': False,
                                'type': bool,
-                               'action': 'store_true'}
+                               'action': 'store_const',
+                               'const': True}
 
 STANDARD_CONFIGS['config_fn'] = {'default': None,
                                  'type': str}
@@ -162,18 +164,21 @@ class Config(object):
     def __init__(self,
                  app_name: str,
                  short_help: str,
-                 long_help: str) -> None:
+                 long_help: str,
+                 validate_dialect: bool=True) -> None:
 
         self.app_name = splitext(basename(app_name))[0]
         self.short_help = short_help
         self.long_help = long_help
+        self.validate_dialect = validate_dialect
         self.obsolete_options: Dict[str, Any] = {}
         self._app_metadata: META_CONFIG_TYPE = {}
         self.config: Dict = {}
         self.nconfig = self.NConfig
 
 
-    def extend_config(self):
+    def extend_config(self,
+                      override_filename=None):
         """ Provide calling programs function placeholder for add-ons
 
         This function exists so that additional attributes beyond what were picked up from
@@ -182,6 +187,7 @@ class Config(object):
         """
         pass
 
+
     def define_user_config(self):
         """ Provide calling programs function placeholder for defining config
 
@@ -190,11 +196,21 @@ class Config(object):
         pass
 
 
-    def get_config(self) -> Tuple[Any, Dict[Any, Any]]:
+    def define_obsolete_config(self):
+        """ Provide calling programs function placeholder for defining obsolete config
+
+        This is where the calling program defines all obsolete config items.
+        """
+        pass
+
+
+    def get_config(self,
+                   override_filename: Optional[str]=None) -> Tuple[Any, Dict[Any, Any]]:
         self.define_user_config()
+        self.define_obsolete_config()
         self.process_configs()
 
-        self.extend_config()
+        self.extend_config(override_filename)
 
         if self.nconfig.verbosity == 'debug':
             self.print_config(key='dialect')
@@ -253,7 +269,6 @@ class Config(object):
         self.add_standard_metadata('help')
 
 
-
     def process_configs(self,
                         test_cli_args: List = None):
 
@@ -269,7 +284,7 @@ class Config(object):
 
         config_fn = cli_args.get('config_fn', None) or env_args.get('config_fn', None)
         if config_fn:
-            file_args_manager = _FileArgs(config_fn)
+            file_args_manager = _FileArgs(config_fn, self._app_metadata)
             file_config = file_args_manager.file_gristle_app_args
         else:
             file_config = {}
@@ -297,13 +312,15 @@ class Config(object):
 
 
     def generate_config_file(self):
-        with open(self.nconfig.gen_config_fn, 'w') as outbuf:  # pylint: disable=no-member
-            if self.nconfig.gen_config_fn.endswith('.yml'):
+        with open(self.nconfig.gen_config_fn, 'wt') as outbuf:  # pylint: disable=no-member
+            if self.nconfig.gen_config_fn.endswith('.yml') or self.nconfig.gen_config_fn.endswith('.yaml'):
                 filtered_config = {k:v for k,v in self.config.items() if k not in ARG_ONLY_CONFIGS}
                 filtered_config = {k:v for k,v in filtered_config.items() if k != 'gen_config_fn'}
-                yaml.dump(filtered_config, outbuf, indent=4)
+                yaml.dump(filtered_config, outbuf, indent=4, default_flow_style=False)
             elif self.nconfig.gen_config_fn.endswith('.json'):
                 json.dump(self.config, outbuf)
+            else:
+                json.dump(self.config, outbuf)
 
 
     def transform_config(self, config):
@@ -346,14 +363,23 @@ class Config(object):
         self.replace_configs(new_config)
 
 
-    def generate_csv_dialect_config(self):
+    def generate_csv_dialect_config(self,
+                                    override_filename: Optional[str]=None):
         """ Adds the csv dialect to the config.
 
             Added by calling programs within the extend_config method.
         """
+        if override_filename:
+            filenames = [override_filename]
+        else:
+            filenames = self.config['infiles']
         md = self._app_metadata
-        autodetected = csvhelper.get_dialect(infiles=self.config['infiles'],
-                                             verbosity=self.config['verbosity'])
+        try:
+            autodetected = csvhelper.get_dialect(filenames,
+                                                 verbosity=self.config['verbosity'])
+        except FileNotFoundError:
+            comm.abort('Error: File not found when generating csv dialect config',
+                       f"One of these files was not found: {','.join(self.config['infiles'])}")
 
         # First override auto-detected dialect with any explicit options
         overridden = csvhelper.override_dialect(autodetected,
@@ -381,19 +407,22 @@ class Config(object):
         self.update_config('dialect', defaulted)
 
 
-    def generate_csv_header_config(self):
+    def generate_csv_header_config(self,
+                                   override_filename=None):
         """ Adds the csv header to the config.
 
             Added by calling programs within the extend_config method.
         """
-
+        if override_filename:
+            filenames = [override_filename]
+        else:
+            filenames = self.config['infiles']
         header = csvhelper.Header()
-        if self.config['infiles'][0] != '-':
-            header.load_from_files(self.config['infiles'], self.config['dialect'])
+        if filenames[0] != '-':
+            header.load_from_files(filenames, self.config['dialect'])
         self.update_config('header', header)
 
 
-
     def _validate_config_metadata(self):
         """ Validates the program's configuration metadata (not the user's input).
         """
@@ -414,6 +443,7 @@ class Config(object):
                 if property_name == 'short_name':
                     if len(property_value) != 1:
                         raise ValueError(f'{arg}.short_name length is invalid')
+
                 if property_name == 'dest':
                     if property_value not in self._app_metadata:
                         raise ValueError(f'{arg}.dest refers to non-existing option')
@@ -427,32 +457,54 @@ class Config(object):
         """ Consolidates environmental and cli arguments.
 
         First all _app_metadata keys are added,
-        Then values from matching environmental variable keys are overlaid,
-        Finally values from matching cli arg keys are overlaid,
+        Then values from matching file, then env, then cli keys are overlaid
         """
         consolidated_args: Dict[str, Any] = {}
 
-        for key in self._app_metadata:
-            actual_key = self._app_metadata[key].get('dest', key)
-            consolidated_args[actual_key] = None
+        def _get_actual_value(key, config_val):
+            bool_val = None
+            if key != self._app_metadata[key].get('dest', key):
+                if (self._app_metadata[key]['type'] is bool
+                        and self._app_metadata[key]['action'] == 'store_const'):
+                    bool_val = self._app_metadata[key]['const']
+                    return bool_val
+            return config_val
 
-        for key, val in file_args.items():
-            try:
+        def _init_dictionary():
+            for key in self._app_metadata:
                 actual_key = self._app_metadata[key].get('dest', key)
-                consolidated_args[actual_key] = val
-            except KeyError:
-                if key in self.obsolete_options.keys():
-                    comm.abort('Error: obsolete option', self.obsolete_options[key])
-                else:
-                    comm.abort(f'ERROR: Unknown option: {key}')
+                consolidated_args[actual_key] = None
+            return consolidated_args
 
-        for key, val in env_args.items():
-            actual_key = self._app_metadata[key].get('dest', key)
-            consolidated_args[actual_key] = val
+        def _add_file_args():
+            for key, val in file_args.items():
+                try:
+                    actual_key = self._app_metadata[key].get('dest', key)
+                    actual_val = _get_actual_value(key, val)
+                    consolidated_args[actual_key] = actual_val
+                except KeyError:
+                    if key in self.obsolete_options.keys():
+                        comm.abort('Error: obsolete option', self.obsolete_options[key])
+                    else:
+                        comm.abort(f'ERROR: Unknown option: {key}')
+            return consolidated_args
 
-        for key, val in cli_args.items():
-            if val is not None and val != []:
-                consolidated_args[key] = val
+        def _add_env_args():
+            for key, val in env_args.items():
+                actual_key = self._app_metadata[key].get('dest', key)
+                consolidated_args[actual_key] = _get_actual_value(key, val)
+            return consolidated_args
+
+        def _add_cli_args():
+            for key, val in cli_args.items():
+                if val is not None and val != []:
+                    consolidated_args[key] = val
+            return consolidated_args
+
+        consolidated_args = _init_dictionary()
+        consolidated_args = _add_file_args()
+        consolidated_args = _add_env_args()
+        consolidated_args = _add_cli_args()
 
         return consolidated_args
 
@@ -526,7 +578,10 @@ class Config(object):
 
 
     def _validate_dialect_with_stdin(self, config) -> None:
-        if config['infiles'] == '-':
+        if self.validate_dialect is False:
+            return
+
+        if config.get('infiles', '') == ['-']:
             if config['delimiter'] is None:
                 comm.abort('Error: csv dialect delimiter is required when piping data via stdin')
             if config['quoting'] is None:
@@ -552,7 +607,6 @@ class Config(object):
                      consolidated_args=None,
                      key=None) -> None:
 
-        print('Config contents: ')
         for a_key in self.config.keys():
             if key and key != a_key:
                 continue
@@ -617,7 +671,7 @@ class _EnvironmentalArgs(object):
                  app_config: METADATA_TYPE) -> None:
 
         self.gristle_app_name = gristle_app_name
-        self._app_metadata = app_config
+        self.app_metadata = app_config
         self.env_args = os.environ.items()
         self.env_gristle_app_args = self._get_gristle_app_args()
 
@@ -630,21 +684,15 @@ class _EnvironmentalArgs(object):
         for envkey, envval in sorted(self.env_args):
             if envkey in self._transform_config_keys_to_env_keys():
                 option_name = self._transform_env_key_to_option_name(envkey)
-                if self._app_metadata[option_name]['type'] is bool:
-                    if self._app_metadata[option_name].get('dest', option_name) == option_name:
-                        # issue: assumes env true/false matches app_config 'store_true'/'store_false'
-                        env_config[option_name] = self._app_metadata[option_name]['type'](envval)
-                    else:
-                        pass # ignore booleans used to reverse a different boolean
-                else:
-                    env_config[option_name] = self._app_metadata[option_name]['type'](envval)
-        return env_config
+                env_config[option_name] = self.app_metadata[option_name]['type'](envval)
+        env_config_binaries_cleaned = binary_arg_fixer(self.app_metadata, env_config)
+        return env_config_binaries_cleaned
 
     def _transform_config_keys_to_env_keys(self):
         """ Translates the configuration keys to the env key formats by adding
             'gristle_[app_name] as a prefix to each key.
         """
-        env_keys = [f'{self.gristle_app_name}_{key}' for key in self._app_metadata.keys()]
+        env_keys = [f'{self.gristle_app_name}_{key}' for key in self.app_metadata.keys()]
         return env_keys
 
     def _transform_env_key_to_option_name(self,
@@ -660,8 +708,7 @@ class _EnvironmentalArgs(object):
 class _FileArgs(object):
     """ Manages all file-based configuration arguments related to the gristle app
 
-        This class will collect these arguments from a config file,
-        format them,
+        This class will collect these arguments from a config file, format them,
         then provide them via class attributes:
            self.file_gristle_app_args
 
@@ -676,15 +723,20 @@ class _FileArgs(object):
                    c.  outfiles
                    d.  outdir
                    e.  tmpdir
+            3. Any keys that are binaries will be validated to ensure they evaluate
+               to true, and then their const values will be applied to their
+               destinations.
     """
 
     def __init__(self,
-                 config_fn: str) -> None:
+                 config_fn: str,
+                 app_metadata) -> None:
 
         if not isfile(config_fn):
             comm.abort(f'Error: config file not found!',
                        f'for config-fn: {config_fn}')
         self.config_fn = config_fn
+        self.app_metadata = app_metadata
         self.file_gristle_app_args = self._get_args()
 
 
@@ -699,19 +751,31 @@ class _FileArgs(object):
         _, file_ext = os.path.splitext(self.config_fn)
         if file_ext in ('.yaml', '.yml'):
             with open(self.config_fn) as buf:
-                file_args = yaml.safe_load(buf)
+                file_args = yaml.YAML(typ='safe', pure=True).load(buf)
         elif file_ext in ('.json'):
             with open(self.config_fn) as buf:
                 file_args = json.load(buf)
 
-        self._convert_arg_name_delimiter(file_args)
-        self._convert_file_path('infiles', file_args)
-        self._convert_file_path('outfile', file_args)
-        self._convert_file_path('outfiles', file_args)
-        self._convert_file_path('outdir', file_args)
-        self._convert_file_path('out_dir', file_args)
-        self._convert_file_path('tmpdir', file_args)
-        return file_args
+        file_args_binaries_cleaned = binary_arg_fixer(self.app_metadata, file_args)
+        file_args_final = self.file_args_path_cleaner(file_args_binaries_cleaned)
+        return file_args_final
+
+
+    def file_args_path_cleaner(self,
+                               file_args):
+
+        output_args = copy.deepcopy(file_args)
+
+        self._convert_arg_name_delimiter(output_args)
+        self._convert_file_path('infiles', output_args)
+        self._convert_file_path('outfile', output_args)
+        self._convert_file_path('outfiles', output_args)
+        self._convert_file_path('outdir', output_args)
+        self._convert_file_path('out_dir', output_args)
+        self._convert_file_path('tmpdir', output_args)
+        self._convert_file_path('source_dir', output_args)
+        self._convert_file_path('dest_dir', output_args)
+        return output_args
 
 
     def _convert_file_path(self, path_key, args):
@@ -870,3 +934,58 @@ class _CommandLineArgs(object):
         if known_args.version:
             print(__version__)
             sys.exit(0)
+
+
+
+def binary_arg_fixer(app_metadata,
+                     args):
+    """ Returns a copy of the args in which:
+        - keys are replaced by any destination key names if they exist
+        - and store_const is used rather than value for bools
+
+        The reason for this function is that unlike cli args, envvars
+        and config files don't have flags whose mere existance indicates
+        True or False: instead config files have bools which can be set
+        either way and envvars just have strings - which might have a True
+        or False, 1 or 0.  This code helps treat the envvars & config files
+        like cli args.
+    """
+
+    def get_bool_actual_value(key,
+                              orig_config_val):
+        assert app_metadata[key]['type'] is bool
+        assert app_metadata[key]['action'] == 'store_const'
+        assert app_metadata[key]['const'] in (True, False)
+
+        if type(orig_config_val) is bool:
+            transformed_config_val = orig_config_val
+        elif orig_config_val is None or orig_config_val.strip().lower() in ('none', 'null'):
+            comm.abort(f'Config item {key} has a non-true value of {orig_config_val}',
+                       'This is a flag type whose value is established by pgm metadata '
+                       'and when provided via envvar or config file must always be set to true')
+        else:
+            transformed_config_val = orig_config_val.strip().lower() in ('true', 't', '1', '')
+
+        if not transformed_config_val:
+            comm.abort(f'Config item {key} has a non-true value of {orig_config_val}',
+                       'This is a flag type whose value is established by pgm metadata'
+                       'and when provided via envvar or config file must always be set to true')
+
+        return app_metadata[key]['const']
+
+    cleaned_args = {}
+    for orig_key, val in args.items():
+        try:
+            if app_metadata[orig_key]['type'] != bool:
+                cleaned_args[orig_key] = val
+            else:
+                actual_key = app_metadata[orig_key].get('dest', orig_key)
+                actual_val = get_bool_actual_value(orig_key, val)
+                cleaned_args[actual_key] = actual_val
+        except KeyError:
+            comm.abort(f'Error: option {orig_key} is unknown')
+
+    return cleaned_args
+
+
+
