@@ -20,23 +20,58 @@ import datagristle.file_io as file_io
 class SortKeysConfig(object):
 
     def __init__(self,
-                 sort_keys: List[str]):
+                 sort_keys: List[str],
+                 header=None):
 
         self.key_fields: List[SortKeyRecord] = []
-        self.load_config(sort_keys)
+        self.load_config(sort_keys, header)
 
 
     def load_config(self,
-                    sort_keys: List[str]) -> None:
+                    sort_keys: List[str],
+                    header: csvhelper.Header = None) -> None:
 
         for key_field in sort_keys:
-            sort_key_rec = SortKeyRecord(key_field)
+            transformed_key_field = self._transform_key_field(header, key_field)
+            sort_key_rec = SortKeyRecord(transformed_key_field)
             self.key_fields.append(sort_key_rec)
 
 
-    def multi_string_orders(self) -> bool:
-        string_orders = set([x.order for x in self.key_fields if x.type == 'str'])
-        return bool(len(string_orders) > 1)
+    def _transform_key_field(self,
+                             header: Optional[csvhelper.Header],
+                             key_field: str) -> str:
+        """key_field looks like:
+               - [field][~][type][~][direction]
+               - ex: 0sf
+               - ex: 3ir
+               - ex: home_statesf
+               - ex: home_state~s~f
+               - ex: 0~s~f
+        """
+        if key_field.count('~') == 2:
+            field, keytype, direction = key_field.split('~')
+        else:
+            direction = key_field[-1]
+            keytype = key_field[-2]
+            field = key_field[:-2]
+
+        if comm.isnumeric(field):
+            field_offset = field
+        else:
+            try:
+                field_offset = str(header.get_field_position(field))
+            except KeyError:
+                comm.abort('Error: field name not found in header',
+                           f'field name: {field}        '
+                           f'header fields: {header.field_names}')
+
+        new_key_field = f'{field_offset}{keytype}{direction}'
+        return new_key_field
+
+
+    def multi_orders(self) -> bool:
+        orders = set([x.order for x in self.key_fields])
+        return bool(len(orders) > 1)
 
 
     def get_primary_order(self) -> str:
@@ -45,9 +80,9 @@ class SortKeysConfig(object):
         """
         string_order = list(set([x.order for x in self.key_fields if x.type == 'str']))
         if string_order:
-            return string_order[0]
+            return 'forward' if 'forward' in string_order else 'reverse'
         else:
-            return self.key_fields[0].order # yep, returning any ole arbitrary type
+            return 'forward' if 'forward' in [x.order for x in self.key_fields] else 'reverse'
 
     def get_sort_fields(self) -> List[int]:
         """ Get the list of columns to sort the self.keys list on
@@ -116,10 +151,11 @@ class CSVPythonSorter(object):
 
         self.all_recs: List[str] = []
         self.keys: List[Tuple[Any, ...]] = []
-        self.header_rec = None
+        self.header_rec: List[str] = []
 
         self.stats = {}
         self.stats['recs_deduped'] = 0
+        self.has_header_adjustment = 0
 
         #todo: handle relative path subtleties:
         #    for reference:  https://stackoverflow.com/questions/918154/relative-paths-in-python
@@ -139,14 +175,14 @@ class CSVPythonSorter(object):
         """
         self._load_file_and_prepare_data()
 
-        if self.sort_key_config.multi_string_orders():
+        if self.sort_key_config.multi_orders():
             self._multipass_sort()
         else:
             self._singlepass_sort()
 
         self._write_file_and_dedupe()
 
-        self.stats['recs_read'] = self.input_handler.rec_cnt
+        self.stats['recs_read'] = self.input_handler.rec_cnt + self.has_header_adjustment
         self.stats['recs_written'] = self.output_handler.rec_cnt
 
 
@@ -167,17 +203,15 @@ class CSVPythonSorter(object):
         all_recs = self.all_recs
         primary_order = self.sort_key_config.get_primary_order()
         if self.input_handler.dialect.has_header:
-            has_header_adjustment = 1
+            self.has_header_adjustment = 1
         else:
-            has_header_adjustment = 0
+            self.has_header_adjustment = 0
 
+        self.header_rec = self.input_handler.header
         for rec in self.input_handler:
-            if self.input_handler.dialect.has_header and self.input_handler.rec_cnt == 1:
-                self.header_rec = rec
-            else:
-                all_recs.append(rec)
-                sort_values = self._get_sort_values(self.sort_key_config.key_fields, rec, primary_order)
-                keys.append((*sort_values, self.input_handler.rec_cnt - 1 - has_header_adjustment))
+            all_recs.append(rec)
+            sort_values = self._get_sort_values(self.sort_key_config.key_fields, rec, primary_order)
+            keys.append((*sort_values, self.input_handler.rec_cnt - 1))
         #print(f'    duration = {(time.time() - start_time):.4f}')
 
 
@@ -189,17 +223,18 @@ class CSVPythonSorter(object):
         try:
             sort_values = [transform(rec[key_field.position], key_field, primary_order) for key_field in key_fields]
         except IndexError:
-            comm.abort('Error: key references columns that does not exist in record', f'{rec=}')
+            comm.abort('Error: key references columns that does not exist in record', f'rec={rec}')
         return sort_values
 
 
     def _singlepass_sort(self) -> None:
-        """ Sorts the keys in a single direction
+        """ Faster sort that has limited ability to support multiple keys and orders
         """
         #print('\n ------------------ Sort Phase: -------------------------')
         start_time = time.time()
         sort_fields = self.sort_key_config.get_sort_fields()
         primary_order = self.sort_key_config.get_primary_order()
+
         if primary_order == 'forward':
             self.keys.sort(key=itemgetter(*sort_fields))
         else:
@@ -208,7 +243,7 @@ class CSVPythonSorter(object):
 
 
     def _multipass_sort(self) -> None:
-        """ Sorts keys in multiple directions
+        """ Slower sort that can handle keys with multiple orders
         """
         #print('\n ------------------ Multi-Pass Sort Phase: -------------------------')
         start_time = time.time()
@@ -224,7 +259,6 @@ class CSVPythonSorter(object):
 
 
     def _write_file_and_dedupe(self) -> None:
-        #note: it is no longer writing the header record out
         #print('\n ------------------ Write Phase: -------------------------')
         keys = self.keys
         all_recs = self.all_recs
@@ -232,7 +266,7 @@ class CSVPythonSorter(object):
 
         start_time = time.time()
 
-        # Run it once to initiate - especially for unit testing.
+        # Run it once to initiate
         isduplicate(None)
 
         if self.keep_header and self.header_rec:
@@ -262,7 +296,8 @@ def isduplicate(key: Tuple[Any, ...],
 def transform(field_value: str,
               key_field: SortKeyRecord,
               primary_order: str) -> Union[str, int, float]:
-
+    """ transforms fields into their type: int, float, str
+    """
     transformed_field_value: Union[str, int, float] = None
     if key_field.type == 'str':
         transformed_field_value = field_value
@@ -273,12 +308,7 @@ def transform(field_value: str,
     else:
         raise ValueError(f'Invalid key-field type: {key_field.type}')
 
-    if (primary_order == 'forward'
-            and key_field.type in ('int', 'float')
-            and key_field.order == 'reverse'):
-        return transformed_field_value * -1
-    else:
-        return transformed_field_value
+    return transformed_field_value
 
 
 
