@@ -32,19 +32,24 @@
        - rpt_collection_analysis_v
 
     See the file "LICENSE" for the full license governing this code.
-    Copyright 2011,2012,2013,2017 Ken Farmer
+    Copyright 2011-2022 Ken Farmer
 """
 
-import appdirs
+import datetime
+import hashlib
+import logging
 import os
-from sqlalchemy import (Table, Column, Boolean, Integer, String, Float,
+from pprint import pprint as pp
+import time
+from typing import Tuple, List, Dict
+
+import appdirs
+from sqlalchemy import (Table, Column, Boolean, Integer, String, Float, Index,
                         MetaData, DATETIME,
                         UniqueConstraint, ForeignKeyConstraint, CheckConstraint,
                         event, text, create_engine)
 from sqlalchemy import exc
-import datetime
-from pprint import pprint as pp
-import logging
+
 import datagristle.simplesql as simplesql
 
 
@@ -155,6 +160,12 @@ class GristleMetaData(object):
         self.field_analysis_value_tools = FieldAnalysisValueTools(self.metadata, self.engine)
         self.field_analysis_value = self.field_analysis_value_tools.table_create()
 
+        self.file_index_tools = FileIndexTools(self.metadata, self.engine)
+        self.file_index = self.file_index_tools.table_create()
+
+        self.migration_tools = MigrationTools(self.metadata, self.engine)
+        self.migration = self.migration_tools.table_create()
+
         self.metadata.create_all()
 
         # we can't easily create views with sqlalchemy - so do that manually:
@@ -248,8 +259,8 @@ class SchemaTools(simplesql.TableTools):
             result = connection.execute(sql,
                                         schema_name=kwargs['schema_name'],
                                         schema_desc=kwargs['schema_desc'])
-        except exc.IntegrityError as e:
-            raise ValueError('Insert failed. %s' % e.message)
+        except exc.IntegrityError as err:
+            raise ValueError('Insert failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -282,8 +293,8 @@ class SchemaTools(simplesql.TableTools):
                                         schema_name=kwargs['schema_name'],
                                         schema_desc=kwargs['schema_desc'],
                                         schema_id=kwargs['schema_id'])
-        except exc.IntegrityError as e:
-            raise ValueError('Update failed. %s' % e.message)
+        except exc.IntegrityError as err:
+            raise ValueError('Update failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -519,9 +530,9 @@ class FieldTools(simplesql.TableTools):
                                         field_order=vkwargs['field_order'],
                                         field_type=vkwargs['field_type'],
                                         field_len=vkwargs['field_len'])
-        except exc.IntegrityError as e:
+        except exc.IntegrityError as err:
             print(sql)
-            raise ValueError('Insert failed. %s' % e.message)
+            raise ValueError('Insert failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -565,9 +576,9 @@ class FieldTools(simplesql.TableTools):
                                         field_type=vkwargs['field_type'],
                                         field_len=vkwargs['field_len'],
                                         element_name=vkwargs['element_name'])
-        except exc.IntegrityError as e:
+        except exc.IntegrityError as err:
             print(sql)
-            raise ValueError('Insert failed. %s' % e.message)
+            raise ValueError('Insert failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -701,8 +712,8 @@ class FieldValueTools(simplesql.TableTools):
                                         fv_value=kwargs['fv_value'],
                                         fv_desc=kwargs['fv_desc'],
                                         fv_issues=kwargs['fv_issues'])
-        except exc.IntegrityError as e:
-            raise ValueError('Insert failed. %s' % e.message)
+        except exc.IntegrityError as err:
+            raise ValueError('Insert failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -734,9 +745,9 @@ class FieldValueTools(simplesql.TableTools):
                                         fv_value=vkwargs['fv_value'],
                                         fv_desc=vkwargs['fv_desc'],
                                         fv_issues=vkwargs['fv_issues'])
-        except exc.IntegrityError as e:
+        except exc.IntegrityError as err:
             print(sql)
-            raise ValueError('Insert failed. %s' % e.message)
+            raise ValueError('Insert failed. %s' % err)
         else:
             return (result.lastrowid,
                     result.rowcount)
@@ -1091,6 +1102,212 @@ class FieldAnalysisValueTools(simplesql.TableTools):
 
 
 
+class FileIndexTools(simplesql.TableTools):
+    """ Includes all methods for the 'file_index' table.
+    """
+
+    def table_create(self):
+        """ Creates the 'file_index' table.
+        """
+        self.file_index = Table('file_index',
+                                self.metadata,
+                                Column('file_hash',
+                                        String(256),
+                                        nullable=False,
+                                        primary_key=True),
+                                Column('record_count',
+                                        Integer,
+                                        nullable=True),
+                                Column('column_count',
+                                        Integer,
+                                        nullable=True),
+                                Column('last_update_epoch',
+                                        Float,
+                                        index=True,
+                                        nullable=False),
+                                UniqueConstraint('file_hash',
+                                                 name='file_index_uk1'),
+                                extend_existing=True)
+
+        self._table = self.file_index
+        self._table_name = 'file_index'
+        self.instance = None # assigned in InstanceTools
+        return self._table
+
+
+    def _hash_file_index(self, filename, mod_datetime, file_bytes):
+        string = bytes(filename + mod_datetime.isoformat() + str(file_bytes), 'utf-8')
+        hash_object = hashlib.sha1(string)
+        hex_digest = hash_object.hexdigest()
+        return hex_digest
+
+
+    def get_file_index_rec_count(self,
+                                 filename,
+                                 mod_datetime,
+                                 file_bytes) -> int:
+        """ Return the record count or -1 if there's no record count
+        """
+
+        file_hash = self._hash_file_index(filename, mod_datetime, file_bytes)
+        sql = """ SELECT record_count
+                  FROM file_index
+                  WHERE file_hash = :file_hash
+              """
+        select_sql = text(sql)
+        result = self.engine.execute(select_sql, file_hash=file_hash)
+        rows = result.fetchall()
+        self.update(filename, mod_datetime, file_bytes)
+        try:
+            return rows[0].record_count
+        except IndexError:  # No rows found
+            return -1
+
+
+    def set_file_index_counts(self,
+                              filename,
+                              mod_datetime,
+                              file_bytes,
+                              rec_count,
+                              col_count) -> Tuple[int, int]:
+        """ Write file_index counts
+        """
+
+        file_hash = self._hash_file_index(filename, mod_datetime, file_bytes)
+        raw_sql = """ INSERT INTO file_index
+                          (file_hash,
+                           record_count,
+                           column_count,
+                           last_update_epoch)
+                      Values(:file_hash,
+                             :rec_count,
+                             :col_count,
+                             :epoch
+                             )
+                  """
+        sql = text(raw_sql)
+        curr_epoch = time.time()
+        try:
+            connection = self.engine.connect()
+            result = connection.execute(sql,
+                                        file_hash=file_hash,
+                                        rec_count=rec_count,
+                                        col_count=col_count,
+                                        epoch=curr_epoch)
+        except exc.IntegrityError as err:
+            raise ValueError('Insert failed. %s' % err)
+        else:
+            self.prune()
+            return (result.lastrowid,
+                    result.rowcount)
+
+
+    def update(self,
+               filename,
+               mod_datetime,
+               file_bytes) -> int:
+        """ Updates the last_update_epoch
+        """
+
+        file_hash = self._hash_file_index(filename, mod_datetime, file_bytes)
+        raw_sql = """ UPDATE file_index
+                      SET last_update_epoch = :epoch
+                      WHERE file_hash = :file_hash
+                  """
+        sql = text(raw_sql)
+        curr_epoch = time.time()
+        try:
+            connection = self.engine.connect()
+            result = connection.execute(sql,
+                                        file_hash=file_hash,
+                                        epoch=curr_epoch)
+        except exc.IntegrityError as err:
+            raise ValueError('Update failed. %s' % err)
+        else:
+            return (result.lastrowid,
+                    result.rowcount)
+
+
+    def prune(self) -> int:
+        """ Deletes any entries more than a year old
+        """
+
+        min_epoch = time.time() - (86400 * 365)
+        raw_sql = """ DELETE FROM file_index
+                      WHERE last_update_epoch < :epoch
+                  """
+        sql = text(raw_sql)
+        try:
+            connection = self.engine.connect()
+            result = connection.execute(sql,
+                                        epoch=min_epoch)
+        except exc.IntegrityError as err:
+            raise ValueError('Delete failed. %s' % err)
+        else:
+            return result.rowcount
+
+
+
+class MigrationTools(simplesql.TableTools):
+    """ Tracks the schema versions
+    """
+
+    def table_create(self):
+        """ Handles creation of collection table.
+        """
+
+        self.migration = Table('migration',
+                                self.metadata,
+                                Column('version',
+                                       String(10),
+                                       nullable=False,
+                                       primary_key=True),
+                                Column('installation_timestamp',
+                                       DATETIME,
+                                       default=datetime.datetime.now,
+                                       onupdate=datetime.datetime.now,
+                                       nullable=False),
+                                Column('comment',
+                                       String(256),
+                                       nullable=True),
+                                extend_existing=True)
+        self._table = self.migration
+        self._table_name = 'migration'
+        return self._table
+
+
+    def write(self):
+        """ Write to migration table
+        """
+        raw_sql = """ DELETE FROM migration"""
+        sql = text(raw_sql)
+        try:
+            connection = self.engine.connect()
+            result = connection.execute(sql)
+        except exc.IntegrityError as err:
+            raise ValueError('Delete failed. %s' % err)
+
+        raw_sql = """ INSERT INTO migration
+                      (version,
+                       installation_timestamp,
+                       comment)
+                      VALUES ('0.2.2',
+                              :dt,
+                              'Adds file_index')
+                  """
+        sql = text(raw_sql)
+        curr_epoch = time.time()
+        try:
+            connection = self.engine.connect()
+            result = connection.execute(sql,
+                                        dt=datetime.datetime.now())
+        except exc.IntegrityError as err:
+            raise ValueError(f'Insert failed: {err}')
+
+
+
+
+
 def create_views(engine, connect):
     """ Creates all views.
         Each view function that it calls is responsible for checking
